#################################################################################
##                                                                             ##
##    Copyright C 2018  Juan P. Dominguez-Morales                              ##
##                                                                             ##
##    This file is part of pyNAVIS.                                            ##
##                                                                             ##
##    pyNAVIS is free software: you can redistribute it and/or modify          ##
##    it under the terms of the GNU General Public License as published by     ##
##    the Free Software Foundation, either version 3 of the License, or        ##
##    (at your option) any later version.                                      ##
##                                                                             ##
##    pyNAVIS is distributed in the hope that it will be useful,               ##
##    but WITHOUT ANY WARRANTY; without even the implied warranty of           ##
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the              ##
##    GNU General Public License for more details.                             ##
##                                                                             ##
##    You should have received a copy of the GNU General Public License        ##
##    along with pyNAVIS.  If not, see <http://www.gnu.org/licenses/>.         ##
##                                                                             ##
#################################################################################


import math
import struct
import csv


class SpikesFile:
    """
    Class that contains all the addresses and timestamps of a file.

    Attributes:
            timestamps (int[]): Timestamps of the file.
            addresses (int[]): Addresses of the file.
    Note:
            Timestamps and addresses are matched, which means that timestamps[0] is the timestamp for the spike with address addresses[0].
    """

    def __init__(self, addresses = [], timestamps = []):
        self.addresses = addresses
        self.timestamps = timestamps

class Loaders:
    """
    Functionalities for loading spiking information from different formats.
    """

    @staticmethod
    def loadAEDAT(path, settings):
        """
        Loads an AEDAT (.aedat) file.
        
        Parameters:
                path (string): Full path of the AEDAT file to be loaded, including name and extension.
                settings (MainSettings): Configuration parameters for the file to load.

        Returns:
                SpikesFile: SpikesFile containing all the addresses and timestamps of the file.
        Raises:
                SettingsError: If settings.address_size is different than 2 and 4.

        """
        unpack_param = ">H"
        
        if settings.address_size == 2:
            unpack_param = ">H"
        elif settings.address_size == 4:
            unpack_param = ">L"
        else:
            print("[Loaders.loadAEDAT] > SettingsError: Only address sizes implemented are 2 and 4 bytes")

        with open(path, 'rb') as f:
            ## Check header ##
            p = 0
            lt = f.readline()
            while lt and lt[0] == ord("#"):
                p += len(lt)
                lt = f.readline()
            f.seek(p)

            f.seek(0, 2)
            eof = f.tell()

            num_events = math.floor((eof-p)/(settings.address_size + 4))

            f.seek(p)

            events = [0] * int(num_events)
            timestamps = [0] * int(num_events)

            ## Read file ##
            i = 0
            try:
                while 1:
                    buff = f.read(settings.address_size)
                    x = struct.unpack(unpack_param, buff)[0]
                    events[i] = x

                    buff = f.read(4)
                    x = struct.unpack('>L', buff)[0]
                    timestamps[i] = x

                    i += 1
            except Exception as inst:
                pass
        spikes_file = SpikesFile([], [])
        spikes_file.addresses = events
        spikes_file.timestamps = timestamps
        return spikes_file


    @staticmethod
    def loadCSV(path):
        """
        Loads a Comma-Separated Values (.csv) file.
        
        Parameters:
                path (string): Full path of the CSV file to be loaded, including name and extension.

        Returns:
                SpikesFile: SpikesFile containing all the addresses and timestamps of the file.

        """
        addresses = []
        timestamps = []
        

        with open(path) as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=',')
            for row in csv_reader:
                addresses.append(int(row[0]))
                timestamps.append(int(row[1]))

        spikes_file = SpikesFile([], [])
        spikes_file.addresses = addresses
        spikes_file.timestamps = timestamps
        return spikes_file


    @staticmethod
    def loadZynqGrabberData(path, settings):
        """
        Loads a text (.txt) file with EAR events collected by the zynqGrabber.
        
        Parameters:
                path (string): Full path of the CSV file to be loaded, including name and extension.
                settings (MainSettings): Configuration parameters for the file to load.

        Returns:
                SpikesFile: SpikesFile containing all the addresses and timestamps of the file.
        """
        
        addresses = []
        timestamps = []
        
        txt_file = open(path, 'r') 
        txt_lines = txt_file.readlines() 
        
        count = 0
        for line in txt_lines:
            event = line.strip().split(',')
            decoded_events_timestamps      = float(event[0])
            decoded_events_auditory_models = int(event[1])           # 0 if the event comes from the NAS
            decoded_events_channels        = int(event[2])           # 0 left, 1 right
            decoded_events_xso_types       = int(event[3])
            decoded_events_neuron_ids      = int(event[4])
            decoded_events_freq_ch_addrs   = int(event[5])
            decoded_events_polarities      = int(event[6])           #0 pos, 1 neg

            if decoded_events_auditory_models == 0: 

                timestamps.append(int(decoded_events_timestamps))
                addresses.append(int(decoded_events_freq_ch_addrs*(1+settings.on_off_both) + decoded_events_polarities +  settings.num_channels*decoded_events_channels*(1+settings.on_off_both)))

        spikes_file = SpikesFile([], [])
        spikes_file.addresses = addresses
        spikes_file.timestamps = timestamps
        return spikes_file


#################################################################################
##                                                                             ##
##    Copyright C 2018  Juan P. Dominguez-Morales                              ##
##                                                                             ##
##    This file is part of pyNAVIS.                                            ##
##                                                                             ##
##    pyNAVIS is free software: you can redistribute it and/or modify          ##
##    it under the terms of the GNU General Public License as published by     ##
##    the Free Software Foundation, either version 3 of the License, or        ##
##    (at your option) any later version.                                      ##
##                                                                             ##
##    pyNAVIS is distributed in the hope that it will be useful,               ##
##    but WITHOUT ANY WARRANTY; without even the implied warranty of           ##
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the              ##
##    GNU General Public License for more details.                             ##
##                                                                             ##
##    You should have received a copy of the GNU General Public License        ##
##    along with pyNAVIS.  If not, see <http://www.gnu.org/licenses/>.         ##
##                                                                             ##
#################################################################################


import copy
import random
import numpy as np
import time
from bisect import bisect_left, bisect_right

from .functions import Functions
from .savers import Savers
from .utils import Utils
from .loaders import SpikesFile

class Splitters:

    @staticmethod
    def manual_splitter(spikes_file, settings, init, end = -1, return_save_both = 0, output_format = '.aedat', path=None):
        """
        Extract a portion of the input SpikesFile file.

        Parameters:
                spikes_file (SpikesFile): Input file.
                init (int): First timestamp from which to start extracting. 
                end (int): Last timestamp from which to stop extracting. 
                settings (MainSettings): Configuration parameters for the input file.
                return_save_both (int, optional): Set it to 0 to return the resultant SpikesFile, to 1 to save the SpikesFile in the output path, and to 2 to do both.
                output_format (string, optional): Output format of the file. Currently supports '.aedat', '.csv', ".txt" and ".txt_rel". See the Savers class for more information.
                path (string, optional): Path where the output file will be saved. Format should not be specified. Not needed if return_save_both is set to 0.

        Returns:
                SpikesFile: SpikesFile containing the extracted portion of the input file. Returned only if return_save_both is either 0 or 2.
        """

        #THIS IS NOT NEEDED IF TS ARE SORTED
        aedat_addr_ts = list(zip(spikes_file.addresses, spikes_file.timestamps))
        aedat_addr_ts = sorted(aedat_addr_ts, key=Utils.getKey)
        spikes_file_new = copy.deepcopy(spikes_file)
        spikes_file_new = Utils.extract_addr_and_ts(aedat_addr_ts)

        if end == -1:
            end = np.max(spikes_file.timestamps)

        a =  bisect_left(spikes_file_new.timestamps, init)
        b =  bisect_right(spikes_file_new.timestamps, end)

        spikes_file_new.addresses = spikes_file_new.addresses[a:b]
        spikes_file_new.timestamps = spikes_file_new.timestamps[a:b]

        if settings.reset_timestamp == True:
            minimum_ts = min(spikes_file_new.timestamps)
            spikes_file_new.timestamps = [(x - minimum_ts) for x in spikes_file_new.timestamps]

        if return_save_both == 0:
            return spikes_file_new
        elif return_save_both == 1 or return_save_both == 2:
            Savers.save_as_any(spikes_file_new, path=path, output_format=output_format, settings=settings) 
            if return_save_both == 2:
                return spikes_file_new


    @staticmethod
    def segmenter_RT(spikes_file, noise_threshold, bin_width, return_save_both = 0, output_format = '.aedat', path=None, settings = None, verbose = False):
        """
        Removes background noise.

        A FIFO of size noise_threshold will be filling up with every spike. If the timestamp difference between the \
             first and the last spike in the FIFO is lower or equal than bin_width, then the most recent spike is saved.

        Parameters:
                spikes_file (SpikesFile): Input file.
                noise_threshold (int): Size of the FIFO. 
                bin_width (int): Time difference (in ms).
                return_save_both (int, optional): Set it to 0 to return the resultant SpikesFile, to 1 to save the SpikesFile in the output path, and to 2 to do both.
                output_format (string, optional): Output format of the file. Currently supports '.aedat', '.csv', ".txt" and ".txt_rel". See the Savers class for more information.
                path (string, optional): Path where the output file will be saved. Format should not be specified. Not needed if return_save_both is set to 0.
                settings (MainSettings, optional): Configuration parameters for the output file. Only needed when saving the output as an AEDAT file.
                verbose (boolean, optional): Set to True if you want the execution time of the function to be printed.

        Returns:
                SpikesFile: SpikesFile containing the processed input file. Returned only if return_save_both is either 0 or 2.
        """

        current_ts = 0  # Timestamp of the current spike that it's being processed.
        spikes_processed = 0    # Number of spikes processed. Used to avoid saving spikes before the FIFO has filled completely.
        buffer_ts = np.zeros(int(noise_threshold))  # FIFO.
        spikes_filtered = SpikesFile([], [])  # SpikesFile where the output will be saved.

        if verbose == True: start_time = time.time()

        for i in range(len(spikes_file.timestamps)):
            current_ts = spikes_file.timestamps[i]
            current_addr = spikes_file.addresses[i]

            buffer_ts = np.roll(buffer_ts, -1)  # Shifting the information of the FIFO.
            buffer_ts[-1] = current_ts

            spikes_processed +=1

            if ((current_ts - buffer_ts[0]) <= bin_width * 1000) and (spikes_processed >= noise_threshold):
                spikes_filtered.addresses.append(current_addr)
                spikes_filtered.timestamps.append(current_ts) 

        if verbose == True: print('SEGMENTER_RT CALCULATION', time.time() - start_time)

        if return_save_both == 0:
            return spikes_filtered
        elif return_save_both == 1 or return_save_both == 2:
            Savers.save_as_any(spikes_filtered, path=path, output_format=output_format) 
            if return_save_both == 2:
                return spikes_filtered


    @staticmethod
    def automatic_splitter(spikes_file, noise_threshold, bin_width, path, settings, output_format = '.aedat'):
        """
        Generate a set of files from the one received as input based on the silences in between words/sounds. This function uses segmenter_RT.
        NOTE: more tests needed to confirm that it works in any case.
        
        Parameters:
                spikes_file (SpikesFile): Input file.                
                noise_threshold (int): Size of the FIFO. 
                bin_width (int): Time difference (in ms).
                path (string): Path where the output files will be saved. Format should not be specified.
                settings (MainSettings, optional): Configuration parameters for the output file. Only needed when saving the output as an AEDAT file.
                output_format (string, optional): Output format of the file. Currently supports '.aedat', '.csv', ".txt" and ".txt_rel". See the Savers class for more information.                
                
        Returns:
                None.
        """
        number_of_splits_generated = 0
        index_previous_split = 0
        new_spikes_file = SpikesFile([], [])
        spikes_filtered =  segmenter_RT(spikes_file,noise_threshold, bin_width, return_save_both = 0)

        for i in range(1, len(spikes_filtered.timestamps)):
            if spikes_filtered.timestamps[i] - spikes_filtered.timestamps[i-1] >= bin_width:
                new_spikes_file.addresses = spikes_filtered.addresses[index_previous_split:i]
                new_spikes_file.timestamps = spikes_filtered.timestamps[index_previous_split:i]
                index_previous_split = i
                number_of_splits_generated += 1

                Savers.save_as_any(new_spikes_file, os.join(path, "split") + str(number_of_splits_generated), output_format, settings)


#################################################################################
##                                                                             ##
##    Copyright C 2018  Juan P. Dominguez-Morales                              ##
##                                                                             ##
##    This file is part of pyNAVIS.                                            ##
##                                                                             ##
##    pyNAVIS is free software: you can redistribute it and/or modify          ##
##    it under the terms of the GNU General Public License as published by     ##
##    the Free Software Foundation, either version 3 of the License, or        ##
##    (at your option) any later version.                                      ##
##                                                                             ##
##    pyNAVIS is distributed in the hope that it will be useful,               ##
##    but WITHOUT ANY WARRANTY; without even the implied warranty of           ##
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the              ##
##    GNU General Public License for more details.                             ##
##                                                                             ##
##    You should have received a copy of the GNU General Public License        ##
##    along with pyNAVIS.  If not, see <http://www.gnu.org/licenses/>.         ##
##                                                                             ##
#################################################################################

import math
import struct
import time
import copy
import random
import datetime

import matplotlib.backends.backend_pdf
import matplotlib.pyplot as plt
import numpy as np

from .loaders import SpikesFile
from .savers import Savers
from .utils import Utils
from .plots import Plots

class Functions:

	@staticmethod
	def check_SpikesFile(spikes_file, settings):
		"""
		Checks if the spiking information contained in the SpikesFile is correct and prints "The loaded SpikesFile file has been checked and it's OK" if the file passes all the checks.
		
		Parameters:
				spikes_file (SpikesFile): File to check.
				settings (MainSettings): Configuration parameters for the file to check.

		Returns:
				None.
		
		Raises:
				TimestampOrderError: If the SpikesFile contains at least one timestamp which value is less than 0.
				TimestampOrderError: If the SpikesFile contains at least one timestamp that is lesser than its previous one.
				AddressValueError: If the SpikesFile contains at least one address less than 0 or greater than the num_channels that you specified in the MainSettings.
		Notes:   
				If mono_stereo is set to 1 (stereo) in the MainSettings, then  addresses should be less than num_channels*2.

				If on_off_both is set to 1 (both) in the MainSettings, then addresses should be less than num_channels*2.
				
				If mono_stereo is set to 1 and on_off_both is set to 1 in the MainSettings, then addresses should be less than num_channels*2*2.
		"""

		if settings.on_off_both == 1:
			number_of_addresses = settings.num_channels*2
		else:
			number_of_addresses = settings.num_channels
		# Check if all timestamps are greater than zero
		a = all(item >= 0  for item in spikes_file.timestamps)

		if not a:
			print("[Functions.check_SpikesFile] > TimestampOrderError: The SpikesFile file that you loaded has at least one timestamp that is less than 0.")

		# Check if each timestamp is greater than its previous one
		b = not any(i > 0 and spikes_file.timestamps[i] < spikes_file.timestamps[i-1] for i in range(len(spikes_file.timestamps)))

		if not b:
			print("[Functions.check_SpikesFile] > TimestampOrderError: The SpikesFile file that you loaded has at least one timestamp whose value is lesser than its previous one.")

		# Check if all addresses are between zero and the total number of addresses
		c = all(item >= 0 and item < number_of_addresses*(settings.mono_stereo + 1) for item in spikes_file.addresses)

		if not c:
			print("[Functions.check_SpikesFile] > AddressValueError: The SpikesFile file that you loaded has at least one event whose address is either less than 0 or greater than the number of addresses that you specified.")

		if a and b and c:
			print("[Functions.check_SpikesFile] > The loaded SpikesFile file has been checked and it's OK")
				

	@staticmethod 
	def adapt_SpikesFile(spikes_file, settings):
		"""
		Subtracts the smallest timestamp of the SpikesFile to all of the timestamps contained in the file (in order to start from 0)
		It also adapts timestamps based on the tick frequency (ts_tick in the MainSettings).
		
		Parameters:
				spikes_file (SpikesFile): File to adapt.
				settings (MainSettings): Configuration parameters for the file to adapt.

		Returns:
				SpikesFile:  Adapted SpikesFile.
		"""
		minimum_ts = min(spikes_file.timestamps)
		if settings.reset_timestamp:
			spikes_file.timestamps = [(x - minimum_ts)*settings.ts_tick for x in spikes_file.timestamps]
		else:
			spikes_file.timestamps = [x*settings.ts_tick for x in spikes_file.timestamps]
		return spikes_file


	@staticmethod
	def phase_lock(spikes_file, settings, posNeg_both = 0):
		"""
		Performs the phase lock operation over a SpikesFile. This can only be performed to SpikeFiles with both ON and OFF addresses. The phaselock operation puts a spike in the output only when the spike train from a specific channel changes from ON (positive part of the signal) to OFF (negative part of the signal). This heavily reduces the number of spikes at the output.
		
		Parameters:
				spikes_file (SpikesFile): File used to perform the phase lock.
				settings (MainSettings): Configuration parameters of the input file.
				posNeg_both (int, optional): If set to 0, a spike is generated only when spike trains change from ON to OFF addresses. If set to 1, a spike is generated every time spike trains change from ON to OFF addresses or vice versa.

		Returns:
				SpikesFile:  Phase-locked SpikesFile.

		Raises:
				SettingsError: If the on_off_both parameter is not set to 1 (both) in the MainSettings.
		"""

		if settings.on_off_both == 1:
			prevSpike = [None] * (settings.num_channels) * (1 + settings.mono_stereo)
			phaseLockedAddrs = []
			phaseLockedTs = []
			for i in range(len(spikes_file.addresses)):
				if prevSpike[spikes_file.addresses[i]//2] == None:
					prevSpike[spikes_file.addresses[i]//2] = spikes_file.addresses[i]%2
				else:
					if (prevSpike[spikes_file.addresses[i]//2] == 0 and spikes_file.addresses[i]%2 == 1) or ((prevSpike[spikes_file.addresses[i]//2] == 1 and spikes_file.addresses[i]%2 == 0) and posNeg_both):
						phaseLockedAddrs.append(spikes_file.addresses[i]//2)
						phaseLockedTs.append(spikes_file.timestamps[i])
						prevSpike[spikes_file.addresses[i]//2] = spikes_file.addresses[i]%2
					else:
						prevSpike[spikes_file.addresses[i]//2] = spikes_file.addresses[i]%2
			spikes_file = SpikesFile([], [])
			spikes_file.addresses = phaseLockedAddrs
			spikes_file.timestamps = phaseLockedTs
			return spikes_file
		else:
			print("[Functions.phase_lock] > SettingsError: this functionality cannot be applied to files that do not have ON/positive and OFF/negative addresses. Check the on_off_both setting for more information.")


	@staticmethod
	def stereo_to_mono(spikes_file, left_right, settings, return_save_both = 0, path = None, output_format = '.aedat'):
		"""
		Generates a mono AEDAT SpikesFile from a stereo SpikesFile.

		Parameters:
				spikes_file (SpikesFile): Input file.
				left_right (int): Set to 0 if you want to extract the left part of the SpikesFile, or to 1 if you want the right part.
				settings (MainSettings): Configuration parameters for the input file.
				return_save_both (int, optional): Set it to 0 to return the SpikesFile, to 1 to save the SpikesFile in the output path, and to 2 to do both.
				path (string, optional): Path where the output file will be saved. Format should not be specified. Not needed if return_save_both is set to 0.
				output_format (string, optional): Output format of the file. Currently supports '.aedat', '.csv', ".txt" and ".txt_rel". See the Savers class for more information.
				

		Returns:
				SpikesFile: SpikesFile containing the shift. Returned only if return_save_both is either 0 or 2.
		
		Raises:
				AttributeError: If the input file is a mono SpikesFile (settings.mono_stereo is set to 0).
		"""

		if settings.mono_stereo:
			addr_ts = list(zip(spikes_file.addresses, spikes_file.timestamps))
			addr_ts = [x for x in addr_ts if x[0] >= left_right*settings.num_channels*(settings.on_off_both + 1) and x[0] < (left_right+1)*settings.num_channels*(settings.on_off_both + 1)]

			spikes_file_mono = Utils.extract_addr_and_ts(addr_ts)
			if left_right:
				spikes_file_mono.addresses = [x-left_right*settings.num_channels*(settings.on_off_both + 1) for x in spikes_file_mono.addresses]
			
			
			if return_save_both == 0:
				return spikes_file_mono
			elif return_save_both == 1 or return_save_both == 2:
				Savers.save_as_any(spikes_file_mono, path=path, output_format=output_format, settings=settings) 
				if return_save_both == 2:
					return spikes_file_mono
			
		else:
			print("[Functions.stereo_to_mono] > SettingsError: this functionality cannot be performed over a mono aedat file.")


	@staticmethod
	def mono_to_stereo(spikes_file, delay, settings, return_save_both = 0, path = None, output_format = '.aedat'):
		"""
		Generates a stereo AEDAT SpikesFile from a mono SpikesFile with a specific delay between both.

		Parameters:
				spikes_file (SpikesFile): Input file.
				delay (int): Delay introduced from left and right spikes. Can be either negative or positive.
				settings (MainSettings): Configuration parameters for the input file.
				return_save_both (int, optional): Set it to 0 to return the SpikesFile, to 1 to save the SpikesFile in the output path, and to 2 to do both.
				path (string, optional): Path where the output file will be saved. Format should not be specified. Not needed if return_save_both is set to 0.
				output_format (string, optional): Output format of the file. Currently supports '.aedat', '.csv', ".txt" and ".txt_rel". See the Savers class for more information.

		Returns:
				SpikesFile: SpikesFile containing the shift. Returned only if return_save_both is either 0 or 2.
		
		Raises:
				SettingsError: If the input file is a stereo SpikesFile (settings.mono_stereo is set to 1).
		"""

		if settings.mono_stereo == 0:
			spikes_file_new = copy.deepcopy(spikes_file)
			newAddrs = [(x + settings.num_channels*(settings.on_off_both+1)) for x in spikes_file_new.addresses]
			spikes_file_new.addresses.extend(newAddrs)
			newTs = [(x + delay) for x in spikes_file_new.timestamps]
			spikes_file_new.timestamps.extend(newTs)
			addr_ts = list(zip(spikes_file_new.addresses, spikes_file_new.timestamps))
			addr_ts = sorted(addr_ts, key=lambda v: (v, random.random())) #key=getKey)  #THIS DISORDERS TSS
			spikes_file_new = Utils.extract_addr_and_ts(addr_ts)

			if delay < 0:
				spikes_file_new.timestamps = [x-delay for x in spikes_file_new.timestamps]

			if return_save_both == 0:
				return spikes_file_new
			elif return_save_both == 1 or return_save_both == 2:
				#settings_new = copy.deepcopy(settings)
				#settings_new.mono_stereo = 1
				Savers.save_as_any(spikes_file_new, path=path, output_format=output_format, settings=settings) 
				if return_save_both == 2:
					return spikes_file_new

		else:
			print("[Functions.mono_to_stereo] > SettingsError: this functionality cannot be performed over a stereo aedat file.")        



	@staticmethod
	def extract_channels_activities(spikes_file, addresses, reset_addresses = True, verbose = False):
		"""
		Extract information from a specific set of addresses from the SpikesFile.
		
		Parameters:
				spikes_file (SpikesFile): File to use.
				addresses (int[]): List of addresses to extract.
				reset_addresses (boolean, optional): If set to true, addresses IDs will start from 0. If not, they will keep their original IDs.
				verbose (boolean, optional): Set to True if you want the execution time of the function to be printed.

		Returns:
				SpikesFile:  SpikesFile containing only the information from the addresses specified as input from spikes_file.
		"""

		if verbose == True: start_time = time.time()
		spikes_per_channels_ts = []
		spikes_per_channel_addr = []
		for i in range(len(spikes_file.timestamps)):
			if spikes_file.addresses[i] in addresses:
				spikes_per_channels_ts.append(spikes_file.timestamps[i])
				spikes_per_channel_addr.append(spikes_file.addresses[i])
		if verbose == True: print('EXTRACT CHANNELS CALCULATION', time.time() - start_time)
		new_spikes_file = SpikesFile([], [])
		new_spikes_file.addresses = spikes_per_channel_addr
		if reset_addresses == True:
			new_spikes_file.addresses = [addr - addresses[0] for addr in new_spikes_file.addresses]
		new_spikes_file.timestamps = spikes_per_channels_ts
		return new_spikes_file


	

	@staticmethod
	def PDF_report(spikes_file, settings, output_path, plots = ["Spikegram", "Sonogram", "Histogram", "Average activity", "Difference between L/R"], vector = False, verbose = False):
		"""
		Generates a PDF report with the spikegram, sonogram, histogram, average activity and difference between L/R plots obtained from the input SpikesFile or path containing SpikeFiles.
		
		Parameters:
				spikes_file (SpikesFile or string): File or path to use.
				settings (MainSettings): Configuration parameters for the input file.
				vector (boolean, optional): Set to True if you want the Spikegram plot vectorized. Note: this may make your PDF heavy.
				verbose (boolean, optional): Set to True if you want the execution time of the function to be printed.

		Returns:
				None.
		
		Notes:   
				If the path used as input is a folder instead of a spikes file, the PDF report is generated for every spikes file contained in the folder.
		"""

		if isinstance(spikes_file, str):
    			
			spikes_file_extension = os.path.splitext(spikes_file)

			if spikes_file_extension == ".aedat":
				spikes_file = Loaders.loadAEDAT(spikes_file, settings)
			elif spikes_file_extension == ".csv":
				spikes_file = Loaders.loadCSV(spikes_file, settings)
			elif spikes_file_extension == ".txt":
				spikes_file = Loaders.loadZynqGrabberData(spikes_file, settings)
    				
			spikes_file = Functions.adapt_SpikesFile(spikes_file, settings)
		
		if isinstance(spikes_file, SpikesFile):
    		
			pdf = matplotlib.backends.backend_pdf.PdfPages(output_path)
			
			# Spikegram
			if any("Spikegram" in s for s in plots):
				spikegram = Plots.spikegram(spikes_file, settings, )
				pdf.savefig(spikegram)
				plt.draw()
			# Sonogram
			if any("Sonogram" in s for s in plots):
				sonogram = Plots.sonogram(spikes_file, settings)		
				pdf.savefig(sonogram)
				plt.draw()
			# Histogram
			if any("Histogram" in s for s in plots):
				Plots.histogram(spikes_file, settings,)		
				pdf.savefig()
				plt.draw()
			# Average activity
			if any("Average activity" in s for s in plots):
				Plots.average_activity(spikes_file, settings,)		
				pdf.savefig()
				plt.draw()
			# Difference between L/R
			if settings.mono_stereo == 1 and any("Difference between L/R" in s for s in plots):		
				Plots.difference_between_LR(spikes_file, settings,)		
				pdf.savefig()
				plt.draw()


			d = pdf.infodict()
			d['Title'] = 'pyNAVIS report'
			d['Author'] = 'Juan P. Dominguez-Morales'
			d['Subject'] = 'pyNAVIS report'
			d['Keywords'] = 'pyNAVIS'
			d['CreationDate'] = datetime.datetime.today()
			d['ModDate'] = datetime.datetime.today()

			pdf.close()

		else:
			print("[Functions.PDF_report] > InputFileError: the input SpikesFile is not valid.")


#################################################################################
##                                                                             ##
##    Copyright C 2018  Juan P. Dominguez-Morales                              ##
##                                                                             ##
##    This file is part of pyNAVIS.                                            ##
##                                                                             ##
##    pyNAVIS is free software: you can redistribute it and/or modify          ##
##    it under the terms of the GNU General Public License as published by     ##
##    the Free Software Foundation, either version 3 of the License, or        ##
##    (at your option) any later version.                                      ##
##                                                                             ##
##    pyNAVIS is distributed in the hope that it will be useful,               ##
##    but WITHOUT ANY WARRANTY; without even the implied warranty of           ##
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the              ##
##    GNU General Public License for more details.                             ##
##                                                                             ##
##    You should have received a copy of the GNU General Public License        ##
##    along with pyNAVIS.  If not, see <http://www.gnu.org/licenses/>.         ##
##                                                                             ##
#################################################################################

import math
import random
from bisect import bisect_left, bisect_right
from matplotlib.colors import LinearSegmentedColormap

import matplotlib.pyplot as plt
import numpy as np
import time

from .utils import Utils

class Plots:
    @staticmethod
    def spikegram(spikes_file, settings, dot_size = 0.2, dot_freq = 1, graph_tile = 'Spikegram', start_at_zero = True, verbose = False):
        """
        Plots the spikegram (also known as cochleogram or raster plot) of a SpikesFile.
        
        This is, a graph where the X axis means time and the Y axis represents addresses (or cochlea channels), and where every spike is plotted as a dot.

        Parameters:
                spikes_file (SpikesFile): File to plot.
                settings (MainSettings): Configuration parameters for the file to plot.
                dot_size (float): Size of the dots used in the spikegram plot.
                dot_freq (int): Set the frequency of spikes that will be represented in the spikegram.
                graph_tile (string, optional): Text that will appear as title for the graph.
                start_at_zero (boolean, optional): If set to True, the X axis will start at 0, instead of starting at the minimum timestamp.
                verbose (boolean, optional): Set to True if you want the execution time of the function to be printed.

        Returns:
                None.

        Note: 
                A value of 10 in dot_size means that for every 10 spikes, only 1 will be plotted. This helps reducing lag when plotting heavy files.
        """

        if verbose == True: start_time = time.time()
        #REPRESENTATION
        plt.style.use('seaborn-whitegrid')
        spk_fig = plt.figure()
        spk_fig.canvas.set_window_title(graph_tile)

        random.seed(0)

        if settings.mono_stereo == 0:
            plt.scatter(spikes_file.timestamps[0::dot_freq], spikes_file.addresses[0::dot_freq], s=dot_size)
        else:
            aedat_addr_ts = list(zip(spikes_file.addresses, spikes_file.timestamps))
            addr, ts = zip(*[(evt[0], evt[1]) for evt in aedat_addr_ts if evt[0] < settings.num_channels*(settings.on_off_both + 1)])
            plt.scatter(ts[0::dot_freq], addr[0::dot_freq], s=dot_size, label='Left cochlea')
            addr, ts = zip(*[(evt[0], evt[1]) for evt in aedat_addr_ts if evt[0] >= settings.num_channels*(settings.on_off_both + 1) and evt[0] < settings.num_channels*(settings.on_off_both + 1)*2])
            plt.scatter(ts[0::dot_freq], addr[0::dot_freq], s=dot_size, label='Right cochlea')
            plt.legend(fancybox=False, ncol=2, loc='upper center', markerscale=2/dot_size, frameon=True)
            
        if verbose == True: print('SPIKEGRAM CALCULATION', time.time() - start_time)

        plt.title(graph_tile, fontsize='x-large')
        plt.xlabel('Timestamp ($\mu$s)', fontsize='large')
        plt.ylabel('Address', fontsize='large')
        plt.ylim([0, settings.num_channels*(settings.on_off_both + 1)*(settings.mono_stereo + 1)])

        if start_at_zero:
            plt.xlim([0, np.max(spikes_file.timestamps)])

        plt.tight_layout()
        spk_fig.show()

    @staticmethod
    def sonogram(spikes_file, settings, return_data = False, graph_tile = 'Sonogram', start_at_zero = True, verbose = False):
        """
        Plots the sonogram of a SpikesFile.
        
        This is, a graph where the X axis means time and the Y axis represents addresses (or cochlea channels), and where the spiking activity is shown with color.

        Parameters:
                spikes_file (SpikesFile): File to plot.
                settings (MainSettings): Configuration parameters for the file to plot.
                return_data (boolean, optional): When set to True, the sonogram matrix will be returned instead of plotted.
                graph_tile (string, optional): Text that will appear as title for the graph.
                start_at_zero (boolean, optional): If set to True, the X axis will start at 0, instead of starting at the minimum timestamp.
                verbose (boolean, optional): Set to True if you want the execution time of the function to be printed.

        Returns:
                int[ , ]: Sonogram matrix. Only returned if return_data is set to True.
        """
        if verbose == True: start_time = time.time()

        if start_at_zero:
            total_time = max(spikes_file.timestamps)
            last_time = 0
        else:
            total_time = max(spikes_file.timestamps) - min(spikes_file.timestamps)
            last_time = min(spikes_file.timestamps)
        

        sonogram = np.zeros((settings.num_channels*(settings.on_off_both + 1)*(settings.mono_stereo+1), int(math.ceil(total_time/settings.bin_size))))

        its = int(math.ceil(total_time/settings.bin_size))
        
        #THIS IS NOT NEEDED IF TS ARE SORTED
        aedat_addr_ts = zip(spikes_file.addresses, spikes_file.timestamps)
        aedat_addr_ts = sorted(aedat_addr_ts, key=Utils.getKey)
        spikes_file = Utils.extract_addr_and_ts(aedat_addr_ts)
        
        for i in range(its):

            a =  bisect_left(spikes_file.timestamps, last_time)
            b =  bisect_right(spikes_file.timestamps, last_time + settings.bin_size)

            blockAddr = spikes_file.addresses[a:b]

            spikes = np.bincount(blockAddr, minlength=settings.num_channels*(settings.on_off_both + 1)*(settings.mono_stereo+1))        

            last_time += settings.bin_size
            sonogram[:, i] = spikes

        if verbose == True: print('SONOGRAM CALCULATION', time.time() - start_time)
        
        if return_data == False:
            # REPRESENTATION
            plt.style.use('default')
            sng_fig = plt.figure()
            sng_fig.canvas.set_window_title(graph_tile)

            plt.imshow(sonogram, aspect="auto", cmap='hot') #, aspect="auto")
            plt.gca().invert_yaxis()

            plt.xlabel('Bin ('+str(settings.bin_size) + '$\mu$s width)', fontsize='large')
            plt.ylabel('Address', fontsize='large')

            plt.title(graph_tile, fontsize='x-large')

            """
            plt.annotate('Right cochlea | Left cochlea',
            xy=(1.00, 0.5), xytext=(5, 0),
            xycoords=('axes fraction', 'figure fraction'),
            textcoords='offset points',
            size=11, ha='center', va='center', rotation=270)
            """

            if settings.mono_stereo == 1:
                plt.annotate('Right cochlea',
                xy=(1.00, 0.75), xytext=(5, 0),
                xycoords=('axes fraction', 'axes fraction'),
                textcoords='offset points',
                size=11, ha='center', va='center', rotation=270)

                plt.annotate('Left cochlea',
                xy=(1.00, 0.25), xytext=(5, 0),
                xycoords=('axes fraction', 'axes fraction'),
                textcoords='offset points',
                size=11, ha='center', va='center', rotation=270)

            colorbar = plt.colorbar()
            colorbar.set_label('No. of spikes', rotation=270, fontsize='large', labelpad= 10)

            sng_fig.show()
        else:
            return sonogram


    @staticmethod
    def histogram(spikes_file, settings, bar_line = 1, graph_tile = 'Histogram', verbose = False):
        """
        Plots the histogram of a SpikesFile.
        
        This is, a graph where addresses (or cochlea channels) are represented in the X axis, and number of spikes in the Y axis.

        Parameters:
                spikes_file (SpikesFile): File to plot.
                settings (MainSettings): Configuration parameters for the file to plot.
                bar_line (int, optional): Select wether to plot the histogram as bar plot (0) or as a line graph (1).
                graph_tile (string, optional): Text that will appear as title for the graph.
                verbose (boolean, optional): Set to True if you want the execution time of the function to be printed.

        Returns:
                int[]: Histogram array.
        """

        start_time = time.time()
        
        spikes_count = np.bincount(spikes_file.addresses, minlength=settings.num_channels * (settings.on_off_both + 1) * (settings.mono_stereo + 1))

        if verbose == True: print('HISTOGRAM CALCULATION:', time.time() - start_time)

        plt.style.use('seaborn-whitegrid')
        hst_fig = plt.figure()
        hst_fig.canvas.set_window_title(graph_tile)
        plt.title(graph_tile, fontsize='x-large')
        plt.xlabel('Address', fontsize='large')
        plt.ylabel('No. of spikes', fontsize='large')

        if bar_line == 0:
            if settings.mono_stereo == 1:
                plt.bar(np.arange(settings.num_channels * (settings.on_off_both + 1)), spikes_count[0:settings.num_channels*(settings.on_off_both + 1)])
                plt.bar(np.arange(settings.num_channels * (settings.on_off_both + 1)), spikes_count[settings.num_channels*(settings.on_off_both + 1):settings.num_channels*2*(settings.on_off_both + 1)])
            else:
                plt.bar(np.arange(settings.num_channels * (settings.on_off_both + 1) * (settings.mono_stereo + 1)), spikes_count)
        else:
            if settings.mono_stereo == 1:
                plt.plot(np.arange(settings.num_channels * (settings.on_off_both + 1)), spikes_count[0:settings.num_channels*(settings.on_off_both + 1)], label='Left cochlea')
                plt.plot(np.arange(settings.num_channels * (settings.on_off_both + 1)), spikes_count[settings.num_channels*(settings.on_off_both + 1):settings.num_channels*2*(settings.on_off_both + 1)], label='Right cochlea')
                plt.legend(loc='best', frameon=True)
            else:
                plt.plot(np.arange(settings.num_channels * (settings.on_off_both + 1) * (settings.mono_stereo + 1)), spikes_count)

        plt.tight_layout()
        hst_fig.show()
        return spikes_count

    @staticmethod
    def average_activity(spikes_file, settings, graph_tile = 'Average activity', verbose=False):
        """
        Plots the average activity plot of a SpikesFile.
        
        This is, a graph where time is represented in the X axis, and average number of spikes in the Y axis.

        Parameters:
                spikes_file (SpikesFile): File to plot.
                settings (MainSettings): Configuration parameters for the file to plot.
                graph_tile (string, optional): Text that will appear as title for the graph.
                verbose (boolean, optional): Set to True if you want the execution time of the function to be printed.

        Returns:
                None.
        """
        aedat_addr_ts = zip(spikes_file.addresses, spikes_file.timestamps)
        total_time = int(max(spikes_file.timestamps))
        last_ts = 0
        average_activity_L = np.zeros(int(math.ceil(total_time/settings.bin_size))+1)
        if(settings.mono_stereo == 1):
            average_activity_R = np.zeros(int(math.ceil(total_time/settings.bin_size))+1)

        #THIS TWO ONLY IF CHECK DETECTS AEDAT NOT IN ORDER
        aedat_addr_ts = sorted(aedat_addr_ts, key=Utils.getKey)
        spikes_file = Utils.extract_addr_and_ts(aedat_addr_ts)

        if verbose == True: start_time = time.time()
        if settings.mono_stereo == 1: 
            for i in range(0, total_time, settings.bin_size):        
                evtL = 0
                evtR = 0

                a =  bisect_left(spikes_file.timestamps, last_ts)
                b =  bisect_right(spikes_file.timestamps, last_ts + settings.bin_size)

                events_list = spikes_file.addresses[a:b]
                
                for j in events_list:
                    if j < settings.num_channels*(1 + settings.on_off_both):
                        evtL = evtL + 1
                    elif j >= settings.num_channels*(1 + settings.on_off_both) and j < settings.num_channels*(1 + settings.on_off_both)*2:
                        evtR = evtR + 1

                average_activity_L[int(i/settings.bin_size)] = evtL
                average_activity_R[int(i/settings.bin_size)] = evtR
                last_ts = last_ts + settings.bin_size
        elif settings.mono_stereo == 0:
            for i in range(0, total_time, settings.bin_size):
                evtL = 0

                a =  bisect_left(spikes_file.timestamps, last_ts)
                b =  bisect_right(spikes_file.timestamps, last_ts + settings.bin_size)
                events_list = spikes_file.addresses[a:b]
                
                for j in events_list:
                    evtL = evtL + 1

                average_activity_L[i//settings.bin_size] = evtL
                last_ts = last_ts + settings.bin_size
        if verbose == True: print('AVERAGE ACTIVITY CALCULATION', time.time() - start_time)

        plt.style.use('seaborn-whitegrid')
        avg_fig = plt.figure()
        avg_fig.canvas.set_window_title(graph_tile)
        plt.title(graph_tile, fontsize='x-large')
        plt.xlabel('Bin ('+str(settings.bin_size) + '$\mu$s width)', fontsize='large')
        plt.ylabel('No. of spikes', fontsize='large')

        plt.plot(np.arange(math.ceil(total_time/settings.bin_size)+1), average_activity_L, label='Left cochlea')
        if(settings.mono_stereo == 1):
            plt.plot(np.arange(math.ceil(total_time/settings.bin_size)+1), average_activity_R, label='Right cochlea')
            plt.legend(loc='best',  ncol=2, frameon=True)

        plt.tight_layout()
        avg_fig.show()

    @staticmethod
    def difference_between_LR(spikes_file, settings, return_data = False, graph_tile = 'Diff. between L and R cochlea', verbose = False):
        """
        Plots a plot showing the differente between the left and the right activity of a SpikesFile.

        Parameters:
                spikes_file (SpikesFile): File to plot.
                settings (MainSettings): Configuration parameters for the file to plot.
                return_data (boolean, optional): When set to True, the sonogram matrix will be returned instead of plotted.
                graph_tile (string, optional): Text that will appear as title for the graph.
                verbose (boolean, optional): Set to True if you want the execution time of the function to be printed.

        Returns:
                int[ , ]: Disparity matrix. Only returned if return_data is set to True.

        Raises:
                SettingsError: if settings.mono_stereo == 0

        Note:
                This function can only be called if the mono_stereo parameter in settings is set to 1.
        """
        if settings.mono_stereo == 1:    
            total_time = max(spikes_file.timestamps) - min(spikes_file.timestamps)
            diff = np.zeros((settings.num_channels*(settings.on_off_both + 1), int(math.ceil(total_time/settings.bin_size))))

            last_time = min(spikes_file.timestamps)

            its = int(math.ceil(total_time/settings.bin_size))

            #THIS IS NOT NEEDED IF TS ARE SORTED
            aedat_addr_ts = list(zip(spikes_file.addresses, spikes_file.timestamps))
            aedat_addr_ts = sorted(aedat_addr_ts, key=Utils.getKey)
            spikes_file = Utils.extract_addr_and_ts(aedat_addr_ts)

            if verbose == True: start_time = time.time()
            for i in range(its):
                a =  bisect_left(spikes_file.timestamps, last_time)
                b =  bisect_right(spikes_file.timestamps, last_time + settings.bin_size)

                blockAddr = spikes_file.addresses[a:b]
                
                spikes = np.bincount(blockAddr, minlength=settings.num_channels*(settings.on_off_both + 1)*(settings.mono_stereo+1))
                
                last_time += settings.bin_size

                diff[:, i] = [x1 - x2 for (x1, x2) in list(zip(spikes[0:settings.num_channels*(settings.on_off_both + 1)], spikes[settings.num_channels*(settings.on_off_both + 1):settings.num_channels*(settings.on_off_both + 1)*2]))]
            if verbose == True: print('DIFF CALCULATION', time.time() - start_time)
            
            if max(abs(np.min(diff)), np.max(diff)) != 0: diff = diff*100/max(abs(np.min(diff)), np.max(diff))

            if return_data == False:
                # REPRESENTATION
                plt.style.use('default')
                sng_fig = plt.figure()
                sng_fig.canvas.set_window_title(graph_tile)
                
                #cmap = 'RdBu'
                colors = [(1, 0.49803921568627450980392156862745, 0.05490196078431372549019607843137), (1, 1, 1), (0.12156862745098039215686274509804, 0.46666666666666666666666666666667, 0.70588235294117647058823529411765)]  # R -> G -> B
                n_bins = [3, 6, 10, 100]  # Discretizes the interpolation into bins
                cmap_name = 'my_list'
                cm = LinearSegmentedColormap.from_list(cmap_name, colors, N=100)

                plt.imshow(diff, vmin=-100, vmax=100, aspect="auto", cmap=cm) #, aspect="auto")
                plt.gca().invert_yaxis()

                plt.xlabel('Bin ('+str(settings.bin_size) + '$\mu$s width)', fontsize='large')
                plt.ylabel('Address', fontsize='large')

                plt.title(graph_tile, fontsize='x-large')

                colorbar = plt.colorbar(ticks=[100, 50, 0, -50, -100], orientation='horizontal')
                colorbar.set_label('Cochlea predominance', rotation=0, fontsize='large', labelpad= 10)
                colorbar.ax.set_xticklabels(['100% L\nCochlea', '50%', '0%\nL==R', '50%', '100% R\nCochlea'])
                colorbar.ax.invert_xaxis()
                sng_fig.show()
            else:
                return diff
        else:
            #print("This functionality is only available for stereo AEDAT files.")
            print("[Plots.difference_between_LR] > SettingsError: This functionality is only available for stereo files.")



from .dataset_gen import DatasetGenerators
from .functions import Functions
from .generators import Generators
from .loaders import SpikesFile, Loaders
from .plots import Plots
from .savers import Savers
from .splitters import Splitters
from .utils import Utils
from .main_settings import MainSettings


from setuptools import setup, find_packages

with open('../README.md') as readme_file:
    README = readme_file.read()
"""
with open('HISTORY.md') as history_file:
    HISTORY = history_file.read()
"""
setup_args = dict(
    name='pyNAVIS',
    version='1.0.6',
    description='Useful tools to analyze and process spiking information from neuromorphic auditory sensors.',
    long_description_content_type="text/markdown",
    long_description=README,
    license='GPL',
    packages=find_packages(),
    author='Juan P. Dominguez-Morales',
    author_email='jpdominguez@atc.us.es',
    keywords=['pyNAVIS', 'NAVIS', 'neuromorphic engineering', "neuromorphic audio processing", "neuromorphic sensors"],
    url='https://github.com/jpdominguez/pyNAVIS',
    download_url='https://pypi.org/project/pyNAVIS/'
)

install_requires = [
	'matplotlib',
	'numpy',
	'scipy'
]

if __name__ == '__main__':
    setup(**setup_args, install_requires=install_requires)


#################################################################################
##                                                                             ##
##    Copyright C 2018  Juan P. Dominguez-Morales                              ##
##                                                                             ##
##    This file is part of pyNAVIS.                                            ##
##                                                                             ##
##    pyNAVIS is free software: you can redistribute it and/or modify          ##
##    it under the terms of the GNU General Public License as published by     ##
##    the Free Software Foundation, either version 3 of the License, or        ##
##    (at your option) any later version.                                      ##
##                                                                             ##
##    pyNAVIS is distributed in the hope that it will be useful,               ##
##    but WITHOUT ANY WARRANTY; without even the implied warranty of           ##
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the              ##
##    GNU General Public License for more details.                             ##
##                                                                             ##
##    You should have received a copy of the GNU General Public License        ##
##    along with pyNAVIS.  If not, see <http://www.gnu.org/licenses/>.         ##
##                                                                             ##
#################################################################################

import random

import numpy as np

from .loaders import SpikesFile
from .savers import Savers
from .main_settings import MainSettings

class Generators:
    
    @staticmethod
    def sweep(freq, cycles, num_ch, length, path = None, output_format = '.aedat', return_save_both = 0):
        """
        Generates a SpikesFile with spikes going from address 0 to address N and back, repeating it for a specific number of cycles.
        
        Parameters:
                freq (int): Number of spikes generated per address and half-cycle.
                cycles (int): Number of repetitions.
                num_ch (int): Number of addresses that the sweep will consider.
                length (int): Number of microseconds that the SpikesFile will have.
                path (string, optional): Path where the output file will be saved. Format should not be specified. Not needed if return_save_both is set to 0.
                output_format (string, optional): Output format of the file. Currently supports '.aedat', '.csv', ".txt" and ".txt_rel". See the Savers class for more information.
                return_save_both (int, optional): Set it to 0 to return the SpikesFile, to 1 to save the SpikesFile in the output path, and to 2 to do both.

        Returns:
                SpikesFile: SpikesFile containing the sweep. Returned only if return_save_both is either 0 or 2.
        """

        spikes_file = SpikesFile([], [])
        time_per_cycle = float(length) // cycles
        spikes_per_cycle = freq * (num_ch * 2 - 1)
        time_between_spikes = float(time_per_cycle) // spikes_per_cycle

        addrs = [0] * int(spikes_per_cycle*cycles)
        timestamps = [0] * int(spikes_per_cycle*cycles)
        current_spike_id = 0    

        for c in range(cycles):
            for i in range((num_ch*2)-1):
                for j in range(freq):
                    addrs[current_spike_id] = i - (i//num_ch)*(((i%(num_ch))+1)*2)
                    timestamps[current_spike_id] = int(current_spike_id*time_between_spikes)
                    current_spike_id += 1
        settings = MainSettings(num_channels = num_ch, mono_stereo = 0)

        spikes_file.timestamps = timestamps
        spikes_file.addresses = addrs

        if return_save_both == 0:
            return spikes_file
        elif return_save_both == 1 or return_save_both == 2:
            Savers.save_as_any(spikes_file, path=path, output_format=output_format, settings=settings) 
            if return_save_both == 2:
                return spikes_file

    @staticmethod
    def shift(freq, num_ch, length, path = None, output_format = '.aedat', return_save_both = 0):
        """
        Generates a SpikesFile with spikes going from address 0 to address N.
    
        Parameters:
                freq (int): Number of spikes generated per address.
                num_ch (int): Number of addresses that the shift will consider.
                length (int): Number of microseconds that the SpikesFile will have.
                path (string, optional): Path where the output file will be saved. Format should not be specified. Not needed if return_save_both is set to 0.
                output_format (string, optional): Output format of the file. Currently supports '.aedat', '.csv', ".txt" and ".txt_rel". See the Savers class for more information.
                return_save_both (int, optional): Set it to 0 to return the SpikesFile, to 1 to save the SpikesFile in the output path, and to 2 to do both.

        Returns:
                SpikesFile: SpikesFile containing the shift. Returned only if return_save_both is either 0 or 2.
        """

        spikes_file = SpikesFile([], [])
        total_spikes_no = freq * num_ch
        addrs = [0] * int(total_spikes_no)
        timestamps = [0] * int(total_spikes_no)
        current_spike_id = 0
        
        time_between_spikes = float(length) // total_spikes_no

        for i in range(num_ch):
            for j in range(freq):
                addrs[current_spike_id] = i
                timestamps[current_spike_id] = int(current_spike_id*time_between_spikes)
                current_spike_id += 1
        settings = MainSettings(num_channels = num_ch, mono_stereo = 0)
        spikes_file.timestamps = timestamps
        spikes_file.addresses = addrs
        if return_save_both == 0:
            return spikes_file
        elif return_save_both == 1 or return_save_both == 2:
            Savers.save_as_any(spikes_file, path=path, output_format=output_format, settings=settings) 
            if return_save_both == 2:
                return spikes_file

    @staticmethod
    def random_addrs(freq, num_ch, length, path = None, output_format = '.aedat', return_save_both = 0):
        """
        Generates a SpikesFile with spikes going from address 0 to address N.
        
        Parameters:
                freq (int): Frequency of the spikes generated per address.
                num_ch (int): Number of addresses that the process will consider.
                length (int): Number of microseconds that the SpikesFile will have.
                path (string, optional): Path where the output file will be saved. Format should not be specified. Not needed if return_save_both is set to 0.
                output_format (string, optional): Output format of the file. Currently supports '.aedat', '.csv', ".txt" and ".txt_rel". See the Savers class for more information.
                return_save_both (int, optional): Set it to 0 to return the SpikesFile, to 1 to save the SpikesFile in the output path, and to 2 to do both.

        Returns:
                SpikesFile: SpikesFile containing the spikes. Returned only if return_save_both is either 0 or 2.
        """

        spikes_file = SpikesFile([], [])
        addrs = [0] * int(freq*length/1000000)
        timestamps = [0] * int(freq*length/1000000)
        random.seed(0)
        for i in range(int(freq*length/1000000)):
            addrs[i] = random.randint(0, num_ch-1)
            timestamps[i] = int(i*1000000/freq) #Multiplying by 1000000 to convert timestamps to microseconds
        
        settings = MainSettings(num_channels = num_ch, mono_stereo = 0)
        spikes_file.timestamps = timestamps
        spikes_file.addresses = addrs
        if return_save_both == 0:
            return spikes_file
        elif return_save_both == 1 or return_save_both == 2:
            Savers.save_as_any(spikes_file, path=path, output_format=output_format, settings=settings) 
            if return_save_both == 2:
                return spikes_file


#################################################################################
##                                                                             ##
##    Copyright C 2018  Juan P. Dominguez-Morales                              ##
##                                                                             ##
##    This file is part of pyNAVIS.                                            ##
##                                                                             ##
##    pyNAVIS is free software: you can redistribute it and/or modify          ##
##    it under the terms of the GNU General Public License as published by     ##
##    the Free Software Foundation, either version 3 of the License, or        ##
##    (at your option) any later version.                                      ##
##                                                                             ##
##    pyNAVIS is distributed in the hope that it will be useful,               ##
##    but WITHOUT ANY WARRANTY; without even the implied warranty of           ##
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the              ##
##    GNU General Public License for more details.                             ##
##                                                                             ##
##    You should have received a copy of the GNU General Public License        ##
##    along with pyNAVIS.  If not, see <http://www.gnu.org/licenses/>.         ##
##                                                                             ##
#################################################################################

import os

import matplotlib
import numpy as np

from .functions import Functions
from .loaders import Loaders
from .plots import Plots
from .savers import Savers

class DatasetGenerators:

    @staticmethod
    def generate_sonogram_dataset(path_input_folder, path_output_folder, settings, allow_subdirectories = False, verbose = False):
        """
        Automatically generates and saves sonograms from a set of AEDAT files.
        
        Parameters:
                path_input_folder (string): Path of the folder where AEDAT files are.
                path_output_folder (string): Path of the folder where sonogram images will be saved.
                settings (MainSettings): Configuration parameters of the files in the path_input_folder.
                allow_subdirectories (boolean, optional): Allow the function to navigate deeper in the path_input_folder folder structure to look for input files.
                verbose (boolean, optional): Set to True if you want the execution time of the function to be printed.

        Returns:
                None.
        """

        if verbose == True:
            print("---- SONOGRAM DATASET GENERATION ----")
            start_time = time.time()    
        if not os.path.exists(path_output_folder):
            os.makedirs(path_output_folder)
        files = []
        progress = 0    
        for r, d, f in os.walk(path_input_folder): # r=root, d=directories, f = files
            for file in f:
                if '.aedat' in file:
                    files.append(os.path.join(r, file))
            if allow_subdirectories == False:
                break

        for file in files:
            spikes_file = Loaders.loadAEDAT(file, settings)
            sonogram_data = Plots.sonogram(spikes_file.addresses, spikes_file.timestamps, settings, return_data = True)
            matplotlib.image.imsave(os.path.join(path_output_folder, os.path.basename(file)) + '.png', sonogram_data)
            
            if verbose == True:
                progress += 1
                print("Progress: " + str(float(progress)*100/len(files)) + "%")
        
        if verbose == True: print('Dataset generation process completed. Took ' + str(time.time()-start_time) + " seconds.")


    @staticmethod
    def generate_phaselock_dataset(path_input_folder, path_output_folder, settings, allow_subdirectories = False, verbose = False):
        """
        Automatically generates and saves phaselocked AEDAT files from a set of AEDAT files.
        
        Parameters:
                path_input_folder (string): Path of the folder where AEDAT files are.
                path_output_folder (string): Path of the folder where phaselocked AEDAT files will be saved.
                settings (MainSettings): Configuration parameters of the files in the path_input_folder.
                allow_subdirectories (boolean, optional): Allow the function to navigate deeper in the path_input_folder folder structure to look for input files.
                verbose (boolean, optional): Set to True if you want the execution time of the function to be printed.

        Returns:
                None.
        """

        if verbose == True:
            print("---- PHASELOCK DATASET GENERATION ----")
            start_time = time.time()    
        if not os.path.exists(path_output_folder):
            os.makedirs(path_output_folder)
        files = []
        progress = 0    
        for r, d, f in os.walk(path_input_folder): # r=root, d=directories, f = files
            for file in f:
                if '.aedat' in file:
                    files.append(os.path.join(r, file))
            if allow_subdirectories == False:
                break

        for file in files:
            spikes_file = Loaders.loadAEDAT(file, settings)
            #sonogram_data = Plots.sonogram(spikes_file.addresses, spikes_file.timestamps, settings, return_data = True)
            #matplotlib.image.imsave(os.path.join(path_output_folder, os.path.basename(file)) + '.png', sonogram_data)
            phaselocked_spikes = Functions.phase_lock(spikes_file, settings)
            Savers.save_AEDAT(phaselocked_spikes, os.path.join(path_output_folder, os.path.basename(file)), settings)
            
            if verbose == True:
                progress += 1
                print("Progress: " + str(float(progress)*100/len(files)) + "%")
        
        if verbose == True: print('Dataset generation process completed. Took ' + str(time.time()-start_time) + " seconds.")


    @staticmethod
    def generate_histogram_dataset(path_input_folder, path_output_folder, settings, allow_subdirectories = False, verbose = False):
        """
        Automatically generates and saves histograms in CSV format from a set of AEDAT files.
        
        Parameters:
                path_input_folder (string): Path of the folder where AEDAT files are.
                path_output_folder (string): Path of the folder where CSV files with the histogram information will be saved.
                settings (MainSettings): Configuration parameters of the files in the path_input_folder.
                allow_subdirectories (boolean, optional): Allow the function to navigate deeper in the path_input_folder folder structure to look for input files.
                verbose (boolean, optional): Set to True if you want the execution time of the function to be printed.

        Returns:
                None.
        """

        if verbose == True:
            print("---- HISTOGRAM DATASET GENERATION ----")
            start_time = time.time()    
        if not os.path.exists(path_output_folder):
            os.makedirs(path_output_folder)
        files = []
        progress = 0    
        for r, d, f in os.walk(path_input_folder): # r=root, d=directories, f = files
            for file in f:
                if '.aedat' in file:
                    files.append(os.path.join(r, file))
            if allow_subdirectories == False:
                break

        for file in files:
            spikes_file = Loaders.loadAEDAT(file, settings)            
            histogram_data = Plots.histogram(spikes_file, settings)
            np.savetxt(os.path.join(path_output_folder, os.path.basename(file)) + ".csv", histogram_data, delimiter=",", fmt='%d')
            
            if verbose == True:
                progress += 1
                print("Progress: " + str(float(progress)*100/len(files)) + "%")
        
        if verbose == True: print('Dataset generation process completed. Took ' + str(time.time()-start_time) + " seconds.")


#################################################################################
##                                                                             ##
##    Copyright C 2018  Juan P. Dominguez-Morales                              ##
##                                                                             ##
##    This file is part of pyNAVIS.                                            ##
##                                                                             ##
##    pyNAVIS is free software: you can redistribute it and/or modify          ##
##    it under the terms of the GNU General Public License as published by     ##
##    the Free Software Foundation, either version 3 of the License, or        ##
##    (at your option) any later version.                                      ##
##                                                                             ##
##    pyNAVIS is distributed in the hope that it will be useful,               ##
##    but WITHOUT ANY WARRANTY; without even the implied warranty of           ##
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the              ##
##    GNU General Public License for more details.                             ##
##                                                                             ##
##    You should have received a copy of the GNU General Public License        ##
##    along with pyNAVIS.  If not, see <http://www.gnu.org/licenses/>.         ##
##                                                                             ##
#################################################################################

import math
import struct
import time

import matplotlib.pyplot as plt
import numpy as np

from .loaders import SpikesFile

class Utils:

	@staticmethod
	def extract_addr_and_ts(zipped_addr_ts):
		"""
		Converts a list of [address, timestamp] tuples into a SpikesFile.

		Parameters:
				aedat_addr_ts (list): A list of [address, timestamp] tuples.

		Returns:
				SpikesFile: A SpikesFile object with the addresses and timestamps obtained from zipped_addr_ts
		"""

		spikes_file = SpikesFile([], [])
		spikes_file.addresses = [x[0] for x in zipped_addr_ts]
		spikes_file.timestamps = [x[1] for x in zipped_addr_ts]
		return spikes_file

	@staticmethod
	def execution_time(executing_function, function_params):
		"""
		Calculate the time that a function takes to execute (in seconds).

		Parameters:
				executing_function (function): The name of the function whose execution time wants to be calculated.
				function_params (list): List of the parameters that want to be used in the executing_function.

		Returns:
				float: Time that the function takes to execute.
		"""

		start_time = time.time()
		executing_function(*function_params)
		total_time = time.time() - start_time
		return total_time

	
	@staticmethod
	def getKey(item):
		"""
		Get timestamps. Used to sort zipped list of [address, timestamp] tuples by timestamp.

		Parameters:
				item (tuple): [address, timestamp] tuple.

		Returns:
				int: Timestamp.
		"""

		return item[1]

	@staticmethod
	def order_timestamps(spikes_file):
		"""
		Order the spikes contained in a SpikesFile by timestamp.

		Parameters:
				spikes_file (SpikesFile): Input SpikesFile to order.

		Returns:
				SpikesFile: Ordered SpikesFile.
		"""
	
		aedat_addr_ts = zip(spikes_file.addresses, spikes_file.timestamps)
		aedat_addr_ts = sorted(aedat_addr_ts, key=Utils.getKey)
		spikes_file_ordered = Utils.extract_addr_and_ts(aedat_addr_ts)
		return spikes_file_ordered

	@staticmethod
	def get_info(spikes_file):
		"""
		Prints the number of spikes and the number of microseconds of audio that the SpikesFile contains.
		
		Parameters:
				spikes_file (SpikesFile): File to get the information from.

		Returns:
				None.
		"""

		print("The file contains", len(spikes_file.addresses), "spikes")
		print("The audio has", max(spikes_file.timestamps), 'microsec')